import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:osam/osam.dart';

void main() async {
  final core = Core();
  runApp(MaterialApp(
    home: App(core),
  ));
}

class App extends StatefulWidget {
  final Core core;

  const App(this.core);

  @override
  _AppState createState() => _AppState();
}

class _AppState extends State<App> {
  bool isVisible = true;

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: OsamBuilder(
          stream: widget.core.propertyStream((state) => state.num),
          builder: (ctx, value) {
            return GestureDetector(
              child: Text(value.toString()),
              onTap: () {
                widget.core.update((state) {
                  state.num = state.num + 1;
                  state.news.add(News()..id = state.num);
                });
              },
            );
          },
        ),
      ),
      floatingActionButton: FloatingActionButton(
        onPressed: () {
          widget.core.update((state) {
            state.news
              ..clear()
              ..add(News()..id = 1);
          });
        },
      ),
      body: OsamBuilder<List<News>>(
        stream: widget.core.propertyStream((state) => state.news),
        builder: (ctx, value) {
          return ListView(
            children: value.map((e) {
              return Row(
                children: <Widget>[
                  Text(e.id.toString()),
                  OsamBuilder<bool>(
                    key: ValueKey(e.id),
                    stream: e.propertyStream((state) => state.isFavorite),
                    builder: (ctx, value) {
                      return GestureDetector(
                        child: Icon(value ? Icons.plus_one : Icons.minimize),
                        onTap: () {
                          e.update((state) => state.isFavorite = !value);
                        },
                      );
                    },
                  ),
                ],
              );
            }).toList(),
          );
        },
      ),
    );
  }
}

// ignore: must_be_immutable
class Core extends BaseState<Core> {
  var num = 3;
  var news = <News>[News()..id = 1, News()..id = 2, News()..id = 3];

  @override
  List<Object> get props => [news];
}

// ignore: must_be_immutable
class News extends BaseState<News> {
  bool isFavorite = false;
  int id;

  @override
  List<Object> get props => [id, isFavorite];
}

typedef ValueBuilder<T> = Widget Function(BuildContext context, T value);
typedef ErrorBuilder = Widget Function(BuildContext context, Object error);

class OsamBuilder<T> extends StatefulWidget {
  final ValueStream<T> stream;
  final ValueBuilder<T> builder;
  final ErrorBuilder errorBuilder;

  const OsamBuilder({Key key, @required this.stream, @required this.builder, this.errorBuilder}) : super(key: key);

  @override
  _OsamBuilderState<T> createState() => _OsamBuilderState<T>();
}

class _OsamBuilderState<T> extends State<OsamBuilder<T>> {
  StreamSubscription<T> _subscription;
  T data;
  Object error;

  @override
  void initState() {
    super.initState();
    _subscribe();
  }

  @override
  void didUpdateWidget(OsamBuilder<T> oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (oldWidget.stream != widget.stream) {
      if (_subscription != null) {
        _unsubscribe();
        data = widget.stream.value;
      }
      _subscribe();
    }
  }

  @override
  void dispose() {
    _unsubscribe();
    super.dispose();
  }

  void _subscribe() {
    if (widget.stream != null) {
      _subscription = widget.stream.listen((T event) {
        setState(() {
          data = event;
        });
      }, onError: (Object error) {
        setState(() {
          error = error;
        });
      });
    }
  }

  void _unsubscribe() {
    if (_subscription != null) {
      _subscription.cancel();
      _subscription = null;
    }
  }

  @override
  Widget build(BuildContext context) {
    return error != null ? widget.errorBuilder(context, error) : widget.builder(context, data ?? widget.stream.value);
  }
}
