import 'dart:async';

import 'package:equatable/equatable.dart';
import 'package:osam/osam.dart';
import 'package:osam/src/util/value_stream.dart';

typedef ValueMapper<ST, V> = V Function(ST state);
typedef StateMapper<ST> = void Function(ST state);

// ignore: must_be_immutable
abstract class BaseState<ST> extends Equatable {
  final _stateBroadcaster = StreamController<ST>.broadcast();
  int _lastKnownHashCode;

  ValueStream<ST> get _stateStream => ValueStream(_stateBroadcaster.stream.map((state) => state), this as ST);

  ValueStream<V> propertyStream<V>(ValueMapper<ST, V> mapper) => _stateStream.map(mapper);

  void update(StateMapper<ST> change) {
    _lastKnownHashCode = hashCode;
    try {
      change(this as ST);
    } catch (e) {
      handleError(e);
    }
    if (_lastKnownHashCode != hashCode) {
      _stateBroadcaster.sink.add(this as ST);
    }
  }

  void handleError(Object error) => print(error);

  @override
  List<Object> get props;
}
